@@ -7,6 +7,7 @@
 */
 
 #include "mpg123lib_intern.h"
+#include "icy2utf8.h"
 #include "getbits.h"
 #include "debug.h"
 
@@ -22,6 +23,41 @@
 
 static int initialized = 0;
 
+#define ALIGNCHECK(mh)
+#define ALIGNCHECKK
+/* On compilers that support data alignment but not the automatic stack realignment.
+   We check for properly aligned stack before risking a crash because of badly compiled
+   client program. */
+#if (defined CCALIGN) && (defined NEED_ALIGNCHECK) && ((defined DEBUG) || (defined CHECK_ALIGN))
+
+/* Common building block. */
+#define ALIGNMAINPART \
+	/* minimum size of 16 bytes, not all compilers would align a smaller piece of data */ \
+	double ALIGNED(16) altest[2]; \
+	debug2("testing alignment, with %lu %% 16 = %lu", \
+		(unsigned long)altest, (unsigned long)((size_t)altest % 16)); \
+	if((size_t)altest % 16 != 0)
+
+#undef ALIGNCHECK
+#define ALIGNCHECK(mh) \
+	ALIGNMAINPART \
+	{ \
+		error("Stack variable is not aligned! Your combination of compiler/library is dangerous!"); \
+		if(mh != NULL) mh->err = MPG123_BAD_ALIGN; \
+\
+		return MPG123_ERR; \
+	}
+#undef ALIGNCHECKK
+#define ALIGNCHECKK \
+	ALIGNMAINPART \
+	{ \
+		error("Stack variable is not aligned! Your combination of compiler/library is dangerous!"); \
+		return MPG123_BAD_ALIGN; \
+	}
+
+#endif
+
+
 #ifdef GAPLESS
 /*
 	Take the buffer after a frame decode (strictly: it is the data from frame fr->num!) and cut samples out.
@@ -29,6 +65,9 @@ static int initialized = 0;
 */
 static void frame_buffercheck(mpg123_handle *fr)
 {
+	/* When we have no accurate position, gapless code does not make sense. */
+	if(!fr->accurate) return;
+
 	/* The first interesting frame: Skip some leading samples. */
 	if(fr->firstoff && fr->num == fr->firstframe)
 	{
@@ -63,6 +102,7 @@ static void frame_buffercheck(mpg123_handle *fr)
 
 int attribute_align_arg mpg123_init(void)
 {
+	ALIGNCHECKK
 	if((sizeof(short) != 2) || (sizeof(long) < 4)) return MPG123_BAD_TYPES;
 
 	init_layer2(); /* inits also shared tables with layer1 */
@@ -92,6 +132,17 @@ mpg123_handle attribute_align_arg *mpg123_parnew(mpg123_pars *mp, const char* de
 {
 	mpg123_handle *fr = NULL;
 	int err = MPG123_OK;
+#ifdef DEBUG
+#ifdef CCALIGN
+	double ALIGNED(16) altest[4];
+	if(((size_t)altest) % 16 != 0)
+	{
+		error("Stack variable is not aligned! Your combination of compiler/library is dangerous!");
+		*error = MPG123_BAD_ALIGN;
+		return NULL;
+	}
+#endif
+#endif
 	if(initialized) fr = (mpg123_handle*) malloc(sizeof(mpg123_handle));
 	else err = MPG123_NOT_INITIALIZED;
 	if(fr != NULL)
@@ -133,6 +184,7 @@ mpg123_handle attribute_align_arg *mpg123_parnew(mpg123_pars *mp, const char* de
 int attribute_align_arg mpg123_decoder(mpg123_handle *mh, const char* decoder)
 {
 	enum optdec dt = dectype(decoder);
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 
 	if(dt == nodec)
@@ -167,15 +219,27 @@ int attribute_align_arg mpg123_decoder(mpg123_handle *mh, const char* decoder)
 int attribute_align_arg mpg123_param(mpg123_handle *mh, enum mpg123_parms key, long val, double fval)
 {
 	int r;
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 	r = mpg123_par(&mh->p, key, val, fval);
 	if(r != MPG123_OK){ mh->err = r; r = MPG123_ERR; }
+	else
+	{ /* Special treatment for some settings. */
+#ifdef FRAME_INDEX
+		if(key == MPG123_INDEX_SIZE)
+		{ /* Apply frame index size and grow property on the fly. */
+			r = frame_index_setup(mh);
+			if(r != MPG123_OK) mh->err = MPG123_INDEX_FAIL;
+		}
+#endif
+	}
 	return r;
 }
 
 int attribute_align_arg mpg123_par(mpg123_pars *mp, enum mpg123_parms key, long val, double fval)
 {
 	int ret = MPG123_OK;
+	ALIGNCHECKK
 	if(mp == NULL) return MPG123_BAD_PARS;
 	switch(key)
 	{
@@ -185,9 +249,11 @@ int attribute_align_arg mpg123_par(mpg123_pars *mp, enum mpg123_parms key, long
 		case MPG123_FLAGS:
 #ifndef GAPLESS
 			if(val & MPG123_GAPLESS) ret = MPG123_NO_GAPLESS;
-			else
 #endif
-			mp->flags = val;
+#ifdef FLOATOUT
+			if(val & MPG123_FORCE_8BIT) ret = MPG123_NO_8BIT;
+#endif
+			if(ret == MPG123_OK) mp->flags = val;
 			debug1("set flags to 0x%lx", (unsigned long) mp->flags);
 		break;
 		case MPG123_ADD_FLAGS:
@@ -241,6 +307,13 @@ int attribute_align_arg mpg123_par(mpg123_pars *mp, enum mpg123_parms key, long
 		case MPG123_RESYNC_LIMIT:
 			mp->resync_limit = val;
 		break;
+		case MPG123_INDEX_SIZE:
+#ifdef FRAME_INDEX
+			mp->index_size = val;
+#else
+			ret = MPG123_NO_INDEX;
+#endif
+		break;
 		default:
 			ret = MPG123_BAD_PARAM;
 	}
@@ -250,6 +323,7 @@ int attribute_align_arg mpg123_par(mpg123_pars *mp, enum mpg123_parms key, long
 int attribute_align_arg mpg123_getparam(mpg123_handle *mh, enum mpg123_parms key, long *val, double *fval)
 {
 	int r;
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 	r = mpg123_getpar(&mh->p, key, val, fval);
 	if(r != MPG123_OK){ mh->err = r; r = MPG123_ERR; }
@@ -259,6 +333,7 @@ int attribute_align_arg mpg123_getparam(mpg123_handle *mh, enum mpg123_parms key
 int attribute_align_arg mpg123_getpar(mpg123_pars *mp, enum mpg123_parms key, long *val, double *fval)
 {
 	int ret = 0;
+	ALIGNCHECKK
 	if(mp == NULL) return MPG123_BAD_PARS;
 	switch(key)
 	{
@@ -303,8 +378,33 @@ int attribute_align_arg mpg123_getpar(mpg123_pars *mp, enum mpg123_parms key, lo
 	return ret;
 }
 
+int attribute_align_arg mpg123_getstate(mpg123_handle *mh, enum mpg123_state key, long *val, double *fval)
+{
+	int ret = MPG123_OK;
+	long theval = 0;
+	double thefval = 0.;
+	ALIGNCHECK(mh);
+	if(mh == NULL) return MPG123_ERR;
+
+	switch(key)
+	{
+		case MPG123_ACCURATE:
+			theval = mh->accurate;
+		break;
+		default:
+			mh->err = MPG123_BAD_KEY;
+			ret = MPG123_ERR;
+	}
+
+	if(val  != NULL) *val  = theval;
+	if(fval != NULL) *fval = thefval;
+
+	return ret;
+}
+
 int attribute_align_arg mpg123_eq(mpg123_handle *mh, enum mpg123_channels channel, int band, double val)
 {
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 	if(band < 0 || band > 31){ mh->err = MPG123_BAD_BAND; return MPG123_ERR; }
 	switch(channel)
@@ -322,10 +422,32 @@ int attribute_align_arg mpg123_eq(mpg123_handle *mh, enum mpg123_channels channe
 	return MPG123_OK;
 }
 
+double attribute_align_arg mpg123_geteq(mpg123_handle *mh, enum mpg123_channels channel, int band)
+{
+	double ret = 0.;
+	ALIGNCHECK(mh);
+	if(mh == NULL) return MPG123_ERR;
+
+	/* Handle this gracefully. When there is no band, it has no volume. */
+	if(band > -1 && band < 32)
+	switch(channel)
+	{
+		case MPG123_LEFT|MPG123_RIGHT:
+			ret = 0.5*(REAL_TO_DOUBLE(mh->equalizer[0][band])+REAL_TO_DOUBLE(mh->equalizer[1][band]));
+		break;
+		case MPG123_LEFT:  ret = REAL_TO_DOUBLE(mh->equalizer[0][band]); break;
+		case MPG123_RIGHT: ret = REAL_TO_DOUBLE(mh->equalizer[1][band]); break;
+		/* Default case is already handled: ret = 0 */
+	}
+
+	return ret;
+}
+
 
 /* plain file access, no http! */
-int attribute_align_arg mpg123_open(mpg123_handle *mh, char *path)
+int attribute_align_arg mpg123_open(mpg123_handle *mh, const char *path)
 {
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 
 	mpg123_close(mh);
@@ -335,6 +457,7 @@ int attribute_align_arg mpg123_open(mpg123_handle *mh, char *path)
 
 int attribute_align_arg mpg123_open_fd(mpg123_handle *mh, int fd)
 {
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 
 	mpg123_close(mh);
@@ -344,6 +467,7 @@ int attribute_align_arg mpg123_open_fd(mpg123_handle *mh, int fd)
 
 int attribute_align_arg mpg123_open_feed(mpg123_handle *mh)
 {
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 
 	mpg123_close(mh);
@@ -355,6 +479,7 @@ int attribute_align_arg mpg123_replace_reader( mpg123_handle *mh,
                            ssize_t (*r_read) (int, void *, size_t),
                            off_t   (*r_lseek)(int, off_t, int) )
 {
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 	mh->rdat.r_read = r_read;
 	mh->rdat.r_lseek = r_lseek;
@@ -364,7 +489,9 @@ int attribute_align_arg mpg123_replace_reader( mpg123_handle *mh,
 
 int decode_update(mpg123_handle *mh)
 {
-	long native_rate = frame_freq(mh);
+	long native_rate;
+	ALIGNCHECK(mh);
+	native_rate = frame_freq(mh);
 	debug2("updating decoder structure with native rate %li and af.rate %li", native_rate, mh->af.rate);
 	if(mh->af.rate == native_rate) mh->down_sample = 0;
 	else if(mh->af.rate == native_rate>>1) mh->down_sample = 1;
@@ -440,7 +567,7 @@ static int get_next_frame(mpg123_handle *mh)
 		debug("read frame");
 		mh->to_decode = FALSE;
 		b = read_frame(mh); /* That sets to_decode only if a full frame was read. */
-		debug3("read of frame %li returned %i (to_decode=%i)", (long)mh->num, b, mh->to_decode);
+		debug4("read of frame %li returned %i (to_decode=%i) at sample %li", (long)mh->num, b, mh->to_decode, (long)mpg123_tell(mh));
 		if(b == MPG123_NEED_MORE) return MPG123_NEED_MORE; /* need another call with data */
 		else if(b <= 0)
 		{
@@ -491,6 +618,25 @@ static int get_next_frame(mpg123_handle *mh)
 	return MPG123_OK;
 }
 
+/*
+	Not part of the api. This just decodes the frame and fills missing bits with zeroes.
+	There can be frames that are broken and thus make do_layer() fail.
+*/
+void decode_the_frame(mpg123_handle *fr)
+{
+	size_t needed_bytes = samples_to_bytes(fr, frame_outs(fr, fr->num+1)-frame_outs(fr, fr->num));
+	fr->clip += (fr->do_layer)(fr);
+	/* There could be less data than promised. */
+	if(fr->buffer.fill < needed_bytes)
+	{
+		if(NOQUIET) fprintf(stderr, "Note: broken frame %li, filling up with zeroes\n", (long)fr->num);
+
+		/* One could do a loop with individual samples instead... but zero is zero. */
+		memset(fr->buffer.data + fr->buffer.fill, 0, needed_bytes - fr->buffer.fill);
+		fr->buffer.fill = needed_bytes;
+	}
+}
+
 /*
 	Put _one_ decoded frame into the frame structure's buffer, accessible at the location stored in <audio>, with <bytes> bytes available.
 	The buffer contents will be lost on next call to mpg123_decode_frame.
@@ -505,11 +651,11 @@ static int get_next_frame(mpg123_handle *mh)
 */
 int attribute_align_arg mpg123_decode_frame(mpg123_handle *mh, off_t *num, unsigned char **audio, size_t *bytes)
 {
+	ALIGNCHECK(mh);
 	if(bytes != NULL) *bytes = 0;
 	if(mh == NULL) return MPG123_ERR;
 	if(mh->buffer.size < mh->outblock) return MPG123_NO_SPACE;
 	mh->buffer.fill = 0; /* always start fresh */
-	*bytes = 0;
 	while(TRUE)
 	{
 		/* decode if possible */
@@ -520,17 +666,20 @@ int attribute_align_arg mpg123_decode_frame(mpg123_handle *mh, off_t *num, unsig
 				mh->new_format = 0;
 				return MPG123_NEW_FORMAT;
 			}
-			*num = mh->num;
+			if(num != NULL) *num = mh->num;
 			debug("decoding");
-			mh->clip += (mh->do_layer)(mh);
+
+			decode_the_frame(mh);
+
 			mh->to_decode = mh->to_ignore = FALSE;
 			mh->buffer.p = mh->buffer.data;
 #ifdef GAPLESS
 			/* This checks for individual samples to skip, for gapless mode or sample-accurate seek. */
 			frame_buffercheck(mh);
 #endif
-			*audio = mh->buffer.p;
-			*bytes = mh->buffer.fill;
+			if(audio != NULL) *audio = mh->buffer.p;
+			if(bytes != NULL) *bytes = mh->buffer.fill;
+
 			return MPG123_OK;
 		}
 		else
@@ -540,7 +689,6 @@ int attribute_align_arg mpg123_decode_frame(mpg123_handle *mh, off_t *num, unsig
 			debug1("got next frame, %i", mh->to_decode);
 		}
 	}
-	return MPG123_ERR;
 }
 
 int attribute_align_arg mpg123_read(mpg123_handle *mh, unsigned char *out, size_t size, size_t *done)
@@ -548,6 +696,25 @@ int attribute_align_arg mpg123_read(mpg123_handle *mh, unsigned char *out, size_
 	return mpg123_decode(mh, NULL, 0, out, size, done);
 }
 
+int attribute_align_arg mpg123_feed(mpg123_handle *mh, const unsigned char *in, size_t size)
+{
+	if(mh == NULL) return MPG123_ERR;
+	if(size > 0)
+	{
+		if(in != NULL)
+		{
+			if(feed_more(mh, in, size) != 0) return MPG123_ERR;
+			else return MPG123_OK;
+		}
+		else
+		{
+			mh->err = MPG123_NULL_BUFFER;
+			return MPG123_ERR;
+		}
+	}
+	return MPG123_OK;
+}
+
 /*
 	The old picture:
 	while(1) {
@@ -562,14 +729,18 @@ int attribute_align_arg mpg123_read(mpg123_handle *mh, unsigned char *out, size_
 	}
 */
 
-int attribute_align_arg mpg123_decode(mpg123_handle *mh,unsigned char *inmemory, size_t inmemsize, unsigned char *outmemory, size_t outmemsize, size_t *done)
+int attribute_align_arg mpg123_decode(mpg123_handle *mh, const unsigned char *inmemory, size_t inmemsize, unsigned char *outmemory, size_t outmemsize, size_t *done)
 {
 	int ret = MPG123_OK;
 	size_t mdone = 0;
+	ALIGNCHECK(mh);
 	if(done != NULL) *done = 0;
 	if(mh == NULL) return MPG123_ERR;
-	if(inmemsize > 0)
-	if(feed_more(mh, inmemory, inmemsize) == -1){ ret = MPG123_ERR; goto decodeend; }
+	if(inmemsize > 0 && mpg123_feed(mh, inmemory, inmemsize) != MPG123_OK)
+	{
+		ret = MPG123_ERR;
+		goto decodeend;
+	}
 	if(outmemory == NULL) outmemsize = 0; /* Not just give error, give chance to get a status message. */
 
 	while(ret == MPG123_OK)
@@ -589,7 +760,7 @@ int attribute_align_arg mpg123_decode(mpg123_handle *mh,unsigned char *inmemory,
 				ret = MPG123_NO_SPACE;
 				goto decodeend;
 			}
-			mh->clip += (mh->do_layer)(mh);
+			decode_the_frame(mh);
 			mh->to_decode = mh->to_ignore = FALSE;
 			mh->buffer.p = mh->buffer.data;
 			debug2("decoded frame %li, got %li samples in buffer", (long)mh->num, (long)(mh->buffer.fill / (samples_to_bytes(mh, 1))));
@@ -624,6 +795,7 @@ decodeend:
 long attribute_align_arg mpg123_clip(mpg123_handle *mh)
 {
 	long ret = 0;
+	ALIGNCHECK(mh);
 	if(mh != NULL)
 	{
 		ret = mh->clip;
@@ -647,12 +819,13 @@ static int init_track(mpg123_handle *mh)
 
 int attribute_align_arg mpg123_getformat(mpg123_handle *mh, long *rate, int *channels, int *encoding)
 {
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 	if(init_track(mh) == MPG123_ERR) return MPG123_ERR;
 
-	*rate = mh->af.rate;
-	*channels = mh->af.channels;
-	*encoding = mh->af.encoding;
+	if(rate != NULL) *rate = mh->af.rate;
+	if(channels != NULL) *channels = mh->af.channels;
+	if(encoding != NULL) *encoding = mh->af.encoding;
 	mh->new_format = 0;
 	return MPG123_OK;
 }
@@ -660,6 +833,7 @@ int attribute_align_arg mpg123_getformat(mpg123_handle *mh, long *rate, int *cha
 off_t attribute_align_arg mpg123_timeframe(mpg123_handle *mh, double seconds)
 {
 	off_t b;
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 	b = init_track(mh);
 	if(b<0) return b;
@@ -676,31 +850,39 @@ off_t attribute_align_arg mpg123_timeframe(mpg123_handle *mh, double seconds)
 */
 off_t attribute_align_arg mpg123_tell(mpg123_handle *mh)
 {
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 	if(track_need_init(mh)) return 0;
 	/* Now we have all the info at hand. */
 	debug5("tell: %li/%i first %li buffer %lu; frame_outs=%li", (long)mh->num, mh->to_decode, (long)mh->firstframe, (unsigned long)mh->buffer.fill, (long)frame_outs(mh, mh->num));
 
-	if((mh->num < mh->firstframe) || (mh->num == mh->firstframe && mh->to_decode))
-	{ /* We are at the beginning, expect output from firstframe on. */
-		off_t pos = frame_outs(mh, mh->firstframe);
+	{ /* Funny block to keep C89 happy. */
+		off_t pos = 0;
+		if((mh->num < mh->firstframe) || (mh->num == mh->firstframe && mh->to_decode))
+		{ /* We are at the beginning, expect output from firstframe on. */
+			pos = frame_outs(mh, mh->firstframe);
 #ifdef GAPLESS
-		pos += mh->firstoff;
+			pos += mh->firstoff;
 #endif
-		return SAMPLE_ADJUST(pos);
-	}
-	else if(mh->to_decode)
-	{ /* We start fresh with this frame. Buffer should be empty, but we make sure to count it in.  */
-		return SAMPLE_ADJUST(frame_outs(mh, mh->num) - bytes_to_samples(mh, mh->buffer.fill));
-	}
-	else
-	{ /* We serve what we have in buffer and then the beginning of next frame... */
-		return SAMPLE_ADJUST(frame_outs(mh, mh->num+1) - bytes_to_samples(mh, mh->buffer.fill));
+		}
+		else if(mh->to_decode)
+		{ /* We start fresh with this frame. Buffer should be empty, but we make sure to count it in.  */
+			pos = frame_outs(mh, mh->num) - bytes_to_samples(mh, mh->buffer.fill);
+		}
+		else
+		{ /* We serve what we have in buffer and then the beginning of next frame... */
+			pos = frame_outs(mh, mh->num+1) - bytes_to_samples(mh, mh->buffer.fill);
+		}
+		/* Substract padding and delay from the beginning. */
+		pos = SAMPLE_ADJUST(pos);
+		/* Negative sample offsets are not right, less than nothing is still nothing. */
+		return pos>0 ? pos : 0;
 	}
 }
 
 off_t attribute_align_arg mpg123_tellframe(mpg123_handle *mh)
 {
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 	if(mh->num < mh->firstframe) return mh->firstframe;
 	if(mh->to_decode) return mh->num;
@@ -708,6 +890,14 @@ off_t attribute_align_arg mpg123_tellframe(mpg123_handle *mh)
 	return mh->buffer.fill ? mh->num : mh->num + 1;
 }
 
+off_t attribute_align_arg mpg123_tell_stream(mpg123_handle *mh)
+{
+	ALIGNCHECK(mh);
+	if(mh == NULL) return MPG123_ERR;
+	/* mh->rd is at least a bad_reader, so no worry. */
+	return mh->rd->tell(mh);
+}
+
 static int do_the_seek(mpg123_handle *mh)
 {
 	int b;
@@ -720,8 +910,8 @@ static int do_the_seek(mpg123_handle *mh)
 		mh->to_decode = FALSE;
 		return MPG123_OK;
 	}
-	/*frame_buffers_reset(mh);*/
 	b = mh->rd->seek_frame(mh, fnum);
+debug1("seek_frame returned: %i", b);
 	if(b<0) return b;
 	/* Only mh->to_ignore is TRUE. */
 	if(mh->num < mh->firstframe) mh->to_decode = FALSE;
@@ -731,9 +921,16 @@ static int do_the_seek(mpg123_handle *mh)
 off_t attribute_align_arg mpg123_seek(mpg123_handle *mh, off_t sampleoff, int whence)
 {
 	int b;
-	off_t pos = mpg123_tell(mh); /* adjusted samples */
-debug1("pos=%li", (long)pos);
-	if(pos < 0) return pos; /* mh == NULL is covered in mpg123_tell() */
+	off_t pos;
+	ALIGNCHECK(mh);
+	pos = mpg123_tell(mh); /* adjusted samples */
+	/* pos < 0 also can mean that simply a former seek failed at the lower levels.
+	  In that case, we only allow absolute seeks. */
+	if(pos < 0 && whence != SEEK_SET)
+	{ /* Unless we got the obvious error of NULL handle, this is a special seek failure. */
+		if(mh != NULL) mh->err = MPG123_NO_RELSEEK;
+		return MPG123_ERR;
+	}
 	if((b=init_track(mh)) < 0) return b;
 
 	switch(whence)
@@ -772,9 +969,18 @@ debug1("pos=%li", (long)pos);
 off_t attribute_align_arg mpg123_feedseek(mpg123_handle *mh, off_t sampleoff, int whence, off_t *input_offset)
 {
 	int b;
-	off_t pos = mpg123_tell(mh); /* adjusted samples */
+	off_t pos;
+	ALIGNCHECK(mh);
+	pos = mpg123_tell(mh); /* adjusted samples */
 	debug3("seek from %li to %li (whence=%i)", (long)pos, (long)sampleoff, whence);
+	/* The special seek error handling does not apply here... there is no lowlevel I/O. */
 	if(pos < 0) return pos; /* mh == NULL is covered in mpg123_tell() */
+	if(input_offset == NULL)
+	{
+		mh->err = MPG123_NULL_POINTER;
+		return MPG123_ERR;
+	}
+
 	if((b=init_track(mh)) < 0) return b; /* May need more to do anything at all. */
 
 	switch(whence)
@@ -818,6 +1024,7 @@ off_t attribute_align_arg mpg123_seek_frame(mpg123_handle *mh, off_t offset, int
 {
 	int b;
 	off_t pos = 0;
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 	if((b=init_track(mh)) < 0) return b;
 
@@ -850,21 +1057,33 @@ off_t attribute_align_arg mpg123_seek_frame(mpg123_handle *mh, off_t offset, int
 	return mpg123_tellframe(mh);
 }
 
+int attribute_align_arg mpg123_set_filesize(mpg123_handle *mh, off_t size)
+{
+	ALIGNCHECK(mh);
+	if(mh == NULL) return MPG123_ERR;
+
+	mh->rdat.filelen = size;
+	return MPG123_OK;
+}
+
 off_t attribute_align_arg mpg123_length(mpg123_handle *mh)
 {
 	int b;
 	off_t length;
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 	b = init_track(mh);
 	if(b<0) return b;
 	if(mh->track_samples > -1) length = mh->track_samples;
 	else if(mh->track_frames > 0) length = mh->track_frames*spf(mh);
-	else
+	else if(mh->rdat.filelen >= 0) /* Let the case of 0 length just fall through. */
 	{
 		/* A bad estimate. Ignoring tags 'n stuff. */
 		double bpf = mh->mean_framesize ? mh->mean_framesize : compute_bpf(mh);
 		length = (off_t)((double)(mh->rdat.filelen)/bpf*spf(mh));
 	}
+	else return MPG123_ERR; /* No length info there! */
+
 	length = frame_ins2outs(mh, length);
 #ifdef GAPLESS
 	if(mh->end_os > 0 && length > mh->end_os) length = mh->end_os;
@@ -878,6 +1097,7 @@ int attribute_align_arg mpg123_scan(mpg123_handle *mh)
 	int b;
 	off_t backframe;
 	int to_decode, to_ignore;
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 	if(!(mh->rdat.flags & READER_SEEKABLE)){ mh->err = MPG123_NO_SEEK; return MPG123_ERR; }
 	/* Scan through the _whole_ file, since the current position is no count but computed assuming constant samples per frame. */
@@ -916,6 +1136,7 @@ int attribute_align_arg mpg123_meta_check(mpg123_handle *mh)
 
 int attribute_align_arg mpg123_id3(mpg123_handle *mh, mpg123_id3v1 **v1, mpg123_id3v2 **v2)
 {
+	ALIGNCHECK(mh);
 	if(v1 != NULL) *v1 = NULL;
 	if(v2 != NULL) *v2 = NULL;
 	if(mh == NULL) return MPG123_ERR;
@@ -933,8 +1154,14 @@ int attribute_align_arg mpg123_id3(mpg123_handle *mh, mpg123_id3v1 **v1, mpg123_
 
 int attribute_align_arg mpg123_icy(mpg123_handle *mh, char **icy_meta)
 {
-	*icy_meta = NULL;
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
+	if(icy_meta == NULL)
+	{
+		mh->err = MPG123_NULL_POINTER;
+		return MPG123_ERR;
+	}
+	*icy_meta = NULL;
 
 	if(mh->metaflags & MPG123_ICY)
 	{
@@ -945,20 +1172,36 @@ int attribute_align_arg mpg123_icy(mpg123_handle *mh, char **icy_meta)
 	return MPG123_OK;
 }
 
+char* attribute_align_arg mpg123_icy2utf8(const char* icy_text)
+{
+	return icy2utf8(icy_text);
+}
+
+
 int attribute_align_arg mpg123_index(mpg123_handle *mh, off_t **offsets, off_t *step, size_t *fill)
 {
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
-	if(offsets == NULL || step == NULL || fill == NULL) return MPG123_BAD_INDEX_PAR;
-
+	if(offsets == NULL || step == NULL || fill == NULL)
+	{
+		mh->err = MPG123_BAD_INDEX_PAR;
+		return MPG123_ERR;
+	}
+#ifdef FRAME_INDEX
 	*offsets = mh->index.data;
 	*step    = mh->index.step;
 	*fill    = mh->index.fill;
-
+#else
+	*offsets = NULL;
+	*step    = 0;
+	*fill    = 0;
+#endif
 	return MPG123_OK;
 }
 
 int attribute_align_arg mpg123_close(mpg123_handle *mh)
 {
+	ALIGNCHECK(mh);
 	if(mh == NULL) return MPG123_ERR;
 	if(mh->rd != NULL && mh->rd->close != NULL) mh->rd->close(mh);
 	mh->rd = NULL;
@@ -1005,7 +1248,15 @@ static const char *mpg123_error[] =
 	"Bad parameter handle. (code 25)",
 	"Invalid parameter addresses for index retrieval. (code 26)",
 	"Lost track in the bytestream and did not attempt resync. (code 27)",
-	"Failed to find valid MPEG data within limit on resync. (code 28)"
+	"Failed to find valid MPEG data within limit on resync. (code 28)",
+	"No 8bit encoding possible. (code 29)",
+	"Stack alignment is not good. (code 30)",
+	"You gave me a NULL buffer? (code 31)",
+	"File position is screwed up, please do an absolute seek (code 32)",
+	"Inappropriate NULL-pointer provided.",
+	"Bad key value given.",
+	"There is no frame index (disabled in this build).",
+	"Frame index operation failed."
 };
 
 const char* attribute_align_arg mpg123_plain_strerror(int errcode)
