@@ -6,12 +6,11 @@
 	initially written by Michael Hipp
 */
 
-#include "config.h"
-
-
 #ifndef _MPG123_TERM_H_
 #define _MPG123_TERM_H_
 
+#include "mpg123app.h"
+
 #ifdef HAVE_TERMIOS
 
 #define LOOP_CYCLES	0.500000	/* Loop time in sec */
